#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import sys
import time
import datetime
import socket
from urllib.request import Request, urlopen
from urllib.error import URLError, HTTPError
import base64
import re

# MyDNS.JPのサーバに現在のIPアドレスを通知するスクリプト
# 2017/04/21 version 3.2.0

class MyDns:
    #-----------------------------------------------------------------
    # 以下、環境に合わせて変更する。

    # MyDNS.JPのIPアドレス通知用URL
    MYDNS_URL = 'http://ipv4.mydns.jp/login.html'

    # MyDNS.JPのマスターID
    USER = 'マスターID'

    # MyDNS.JPのパスワード
    PASSWD = 'パスワード'

    # 対象となるドメイン
    TARGET_DOMAIN = 'EXAMPLE.mydns.jp'

    # 現在のIPアドレスを取得するためのURL
    #
    # 返ってくるテキストに所望のIPアドレスが含まれていればよい。
    # ただし、その他のIPアドレスが所望のIPアドレスより先に現れてはならない。
    # 1. `172.217.27.67`のような「IPアドレスのテキストのみ」
    # 2. `REMOTE_ADDR:172.217.27.67`のような「何らかの文字列とIPアドレスのテキスト」
    # 3. `<html><head><title>Current IP Check</title></head>
    #    <body>Current IP Address: 172.217.27.67</body></html>`のような
    #    「HTMLタグ付きでIPアドレスを含むテキスト」
    # 上記のいずれのパターンでもOK
    IP_CHECK_URL = 'https://www.ugtop.com/spill.shtml' # 確認くん
    #IP_CHECK_URL = 'https://env.b4iine.net/' # 確認君+(Plus)
    #IP_CHECK_URL = 'https://env.b4iine.net/isp.php' # プロバイダー確認君

    # IPアドレスに変化がないかチェックする間隔(分単位)
    # 例: 10分に1回チェックなら`10`
    IP_CHECK_INTERVAL = 10

    # IPアドレスに変化があろうがなかろうが定期的に通知する間隔(分単位)
    # 例: 24時間に1回通知なら`24 * 60`
    DEFAULT_INTERVAL  = 24 * 60

    # ログレベル
    # 0: なし: 全てのログを記録しない
    # 1: 通知: 「IPアドレスに変化があった時」「定期通知」「エラー」
    # 2: 全て: 1に加え「IPアドレスに変化がなかった時」も含む
    LOG_LEVEL = 1

    # ログファイルのファイル名
    LOG_FILE_NAME = '/opt/ddnsclient/mydns.log'

    # ここまで
    #-----------------------------------------------------------------

    # Usageメッセージ
    USAGE = '''Usage: python3 {} [-v] [-h]
       -v outputs all messages to stdout
       -h shows usage'''.format(__file__)

    # IPアドレスを表す正規表現
    RE_IP_ADDR = '((?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?))'

    # コンストラクタ
    def __init__(self):
        # インスタンス変数を初期化する。
        self.verbose = False # 標準出力に出力するかどうか
        self.prev_ip_check_datetime = datetime.datetime.today() # 前回IPチェックした時刻
        self.prev_notify_datetime = datetime.datetime.today()   # 前回通知を行った時刻
        self.re_ip_addr = re.compile(self.RE_IP_ADDR) # IPアドレスの正規表現をコンパイルしたもの

        # 通知リクエストを生成しておく。
        # [:-1]は、最後のb'\n'を取るため。
        base64string = base64.encodestring(('%s:%s' % (self.USER, self.PASSWD)).encode('utf-8'))[:-1]
        headers = {
            'Authorization': 'Basic %s' % base64string.decode('utf-8')
        }
        self.notify_req = Request(self.MYDNS_URL, headers=headers)

    # 引数を処理する。
    def handle_args(self):
        # 引数(オプション)を取り出す。
        args = sys.argv
        args.pop(0)
        options = [option for option in args if option.startswith('-')]

        # オプションを処理する。
        if '-h' in options:
            print(self.USAGE)
            return False
        if '-v' in options:
            self.verbose = True
        return True

    # 処理を開始する。
    def start(self):
        # 引数を処理する。
        result = self.handle_args()
        if result == False:
            sys.exit(0)

        # このスクリプトが起動したら、通知しておく。
        self.notify()

        # メインループ
        while True:
            # 登録されているIPアドレスが現在のIPアドレスと異なっていたら直ちに通知する。
            if self.is_ipaddr_changed():
                self.output('IP address changed!', 1)
                self.notify()

            # 一定時間経過したら、定期通知を行う。
            if self.is_default_interval_elapsed():
                self.output('Default interval elapsed.', 1)
                self.notify()

            # 適度にスリープする。
            time.sleep(60) # 60秒

    # 登録されているIPアドレスが現在のIPアドレスと異なっているかチェックする。
    def is_ipaddr_changed(self):
        # チェックのタイミングでなかったら何もしないで終了する。
        elapsed_sec = self.get_elapsed_sec_ip_check()
        if elapsed_sec < self.IP_CHECK_INTERVAL * 60:
            return False

        # 登録されているIPアドレスと現在のIPアドレスを取得する。
        try:
            registered_ip = socket.gethostbyname(self.TARGET_DOMAIN)
            with urlopen(self.IP_CHECK_URL) as res:
                ip_check_str = res.read().decode('utf-8')
                current_ip = self.re_ip_addr.search(ip_check_str).group()
        except socket.gaierror as e:
            self.output('is_ipaddr_changed(): socket.gaierror: {}'.format(e), 1)
            return False
        except socket.herror as e:
            self.output('is_ipaddr_changed(): socket.herror: {}'.format(e), 1)
            return False
        except:
            self.output('is_ipaddr_changed(): Unexpected Error', 1)
            return False

        # IPアドレスチェックを行った時刻を更新する。
        self.prev_ip_check_datetime = datetime.datetime.today()

        # 登録されているIPアドレスが現在のIPアドレスと異なっていたらTrueを返す。
        if registered_ip != current_ip:
            self.output('Registered IP address = {}'.format(registered_ip), 1)
            self.output('Current IP address    = {}'.format(current_ip), 1)
            return True
        else:
            self.output('Registered IP address = {}'.format(registered_ip), 2)
            self.output('Current IP address    = {}'.format(current_ip), 2)
            return False

    # 前回のIPアドレスチェックからの経過秒数を取得する。
    def get_elapsed_sec_ip_check(self):
        current_datetime = datetime.datetime.today()
        timedelta = current_datetime - self.prev_ip_check_datetime

        # 24時間(=86400秒)を越えている場合に備えて、daysも考慮して秒数を算出して返す。
        return timedelta.days * 86400 + timedelta.seconds

    # デフォルトの一定時間が経過したかをチェックする。
    def is_default_interval_elapsed(self):
        current_datetime = datetime.datetime.today()
        timedelta = current_datetime - self.prev_notify_datetime

        # 24時間(=86400秒)を越えている場合に備えて、daysも考慮して秒数を算出する。
        elapsed_sec = timedelta.days * 86400 + timedelta.seconds
        return elapsed_sec >= self.DEFAULT_INTERVAL * 60

    # MyDNS.JPサーバに現在のIPアドレスを通知する。
    def notify(self):
        # 通知のための通信を行う。
        try:
            with urlopen(self.notify_req) as res:
                pass
        except URLError as e:
            self.output('notify(): URLError: {}'.format(e.reason), 1)
            return
        except HTTPError as e:
            self.output('notify(): HTTPError: {}'.format(e.reason), 1)
            return
        except:
            self.output('notify(): Unexpected Error', 1)
            return

        self.output('Notified.', 1)

        # IPアドレスチェックの時刻と通知した時刻を更新する。
        self.prev_ip_check_datetime = datetime.datetime.today()
        self.prev_notify_datetime = datetime.datetime.today()

    # オプションの状態やログレベルにより、標準出力やログファイルにメッセージを出力する。
    def output(self, msg, log_level):
        m = '{}: {}'.format(datetime.datetime.today().strftime('%Y/%m/%d %H:%M:%S'), msg)
        if self.verbose:
            print(m)
        if log_level <= self.LOG_LEVEL:
            self.write_log(m)

    # ログファイルにメッセージを書き出す。
    def write_log(self, msg):
        # ログファイルを開いて、メッセージを追記する。
        try:
            with open(self.LOG_FILE_NAME, 'a') as f:
                f.write('{}\n'.format(msg))
        except OSError as e:
            print('{}: write_log(): {}'.format(__file__, e.strerror))
        except:
            print('{}: write_log(): Unexpected Error'.format(__file__))

# メイン
def main():
    mydns = MyDns()
    mydns.start()

    # NOT REACHED

if __name__ == '__main__':
    main()
